TYPE = "type"
VALUE = "value"

INT = "int"
FLOAT = "float"
BOOL = "bool"
STRING = "str"
COMPLEX = "complex"
NONE_TYPE = "NoneType"

LIST = "list"
TUPLE = "tuple"
SET = "set"
BYTES = "bytes"

DICT = "dict"

FUNCTION = "function"

CLASS = "class"
OBJECT = "object"
CO_NAMES = "co_names"
FIELDS = "__fields__"
OBJECT_TYPE = "__object_type__"

MODULE = "module"

GLOBALS = "__globals__"
CODE = "__code__"
NAME = "__name__"

BUILTINS = "__builtins__"
DOC = "__doc__"
DEFAULTS = "__defaults__"
TRUE = "True"

FUNCTION_ATTRIBUTES = [
    "__code__",
    "__name__",
    "__defaults__"
]

NOT_CLASS_ATTRIBUTES = [
    "__class__",
    "__getattribute__",
    "__new__",
    "__setattr__",
]

CODE_ARGS = [
    'co_argcount',
    'co_posonlyargcount',
    'co_kwonlyargcount',
    'co_nlocals',
    'co_stacksize',
    'co_flags',
    'co_code',
    'co_consts',
    'co_names',
    'co_varnames',
    'co_filename',
    'co_name',
    'co_firstlineno',
    'co_linetable',
    'co_freevars',
    'co_cellvars'
]


import inspect
import types
from types import CodeType, FunctionType
from pydoc import locate

from serializers.source.constants import *


def get_type(item):
    item_type = str(type(item))

    return item_type[8:len(item_type) - 2]


def serialize(item):
    if isinstance(item, (int, float, complex, bool, str, type(None))):
        return serialize_single_var(item)
    elif isinstance(item, (list, tuple, set, bytes)):
        return serialize_collection(item)
    elif isinstance(item, dict):
        return serialize_dict(item)
    elif inspect.isfunction(item):
        return serialize_function(item)
    elif inspect.isclass(item):
        return serialize_class(item)
    elif inspect.iscode(item):
        return serialize_code(item)
    elif inspect.ismodule(item):
        return serialize_module(item)
    elif inspect.ismethoddescriptor(item) or inspect.isbuiltin(item):
        return serialize_instance(item)
    elif inspect.isgetsetdescriptor(item) or inspect.ismemberdescriptor(item):
        return serialize_instance(item)
    elif isinstance(item, type(type.__dict__)):
        return serialize_instance(item)
    else:
        return serialize_object(item)


def serialize_single_var(item):
    serialized_dict = {TYPE: get_type(item), VALUE: item}

    return serialized_dict


def serialize_collection(item):
    serialized_dict = {TYPE: get_type(item), VALUE: [serialize(obj) for obj in item]}

    return serialized_dict


def serialize_dict(item):
    serialized_dict = {TYPE: get_type(item), VALUE: [[serialize(key), serialize(item[key])] for key in item]}

    return serialized_dict


def serialize_function(item):
    members = inspect.getmembers(item)
    serialized = dict()
    serialized['type'] = str(type(item))[8:-2]
    value = dict()

    for tmp in members:
        if tmp[0] in ['__code__', '__name__', '__defaults__']:
            value[tmp[0]] = (tmp[1])
        if tmp[0] == '__code__':
            co_names = tmp[1].__getattribute__('co_names')
            globs = item.__getattribute__('__globals__')
            value['__globals__'] = dict()

            for tmp_co_names in co_names:
                if tmp_co_names == item.__name__:
                    value['__globals__'][tmp_co_names] = item.__name__
                elif not inspect.ismodule(tmp_co_names) \
                        and tmp_co_names in globs:
                    # and tmp_co_names not in __builtins__:
                    value['__globals__'][tmp_co_names] = globs[tmp_co_names]

    serialized['value'] = serialize(value)

    return serialized


def serialize_class(item):
    serialized_dict = {TYPE: CLASS}
    value = {NAME: item.__name__}
    members = inspect.getmembers(item)
    for obj in members:
        if not (obj[0] in NOT_CLASS_ATTRIBUTES):
            value[obj[0]] = obj[1]
    serialized_dict[VALUE] = serialize(value)

    return serialized_dict


def serialize_code(item):
    if get_type(item) is None:
        return None

    members = inspect.getmembers(item)
    serialized_dict = {TYPE: get_type(item), VALUE: serialize({obj[0]: obj[1] for obj in members if not callable(obj[1])})}

    return serialized_dict


def serialize_module(item):
    temp_item = str(item)
    serialized_dict = {TYPE: get_type(item), VALUE: temp_item[9:len(temp_item) - 13]}

    return serialized_dict


def serialize_instance(item):
    members = inspect.getmembers(item)
    serialized_dict = {TYPE: get_type(item), VALUE: serialize({obj[0]: obj[1] for obj in members if not callable(obj[1])})}

    return serialized_dict


def serialize_object(item):
    serialized_dict = {TYPE: OBJECT, VALUE: serialize({OBJECT_TYPE: type(item), FIELDS: item.__dict__})}

    return serialized_dict


def deserialize(item):
    if item[TYPE] in [INT, FLOAT, BOOL, STRING, COMPLEX, NONE_TYPE]:
        return deserialize_single_var(item)
    elif item[TYPE] in [LIST, TUPLE, SET, BYTES]:
        return deserialize_collection(item)
    elif item[TYPE] == DICT:
        return deserialize_dict(item)
    elif item[TYPE] == FUNCTION:
        return deserialize_function(item)
    elif item[TYPE] == CLASS:
        return deserialize_class(item)
    elif item[TYPE] == MODULE:
        return deserialize_module(item)
    elif item[TYPE] == OBJECT:
        return deserialize_object(item)


def deserialize_single_var(item):
    if item[TYPE] == NONE_TYPE:
        return None
    elif item[TYPE] == BOOL and isinstance(item[VALUE], str):
        return item[VALUE] == TRUE
    else:
        return locate(item[TYPE])(item[VALUE])


def deserialize_collection(item):
    if item[TYPE] == LIST:
        return list(deserialize(obj) for obj in item[VALUE])
    elif item[TYPE] == TUPLE:
        return tuple(deserialize(obj) for obj in item[VALUE])
    elif item[TYPE] == SET:
        return set(deserialize(obj) for obj in item[VALUE])
    elif item[TYPE] == BYTES:
        return bytes(deserialize(obj) for obj in item[VALUE])


def deserialize_dict(item):
    return {deserialize(obj[0]): deserialize(obj[1]) for obj in item[VALUE]}


def deserialize_function(item):
    res_dict = deserialize(item['value'])

    res_dict['code'] = deserialize_code(item)
    res_dict.pop('__code__')

    res_dict['globals'] = res_dict['__globals__']
    res_dict.pop('__globals__')

    res_dict['name'] = res_dict['__name__']
    res_dict.pop('__name__')

    res_dict['argdefs'] = res_dict['__defaults__']
    res_dict.pop('__defaults__')

    res = types.FunctionType(**res_dict)
    if res.__name__ in res.__getattribute__('__globals__'):
        res.__getattribute__('__globals__')[res.__name__] = res

    return res


def deserialize_code(item):
    items = item['value']['value']

    for tmp in items:
        if tmp[0]['value'] == '__code__':
            args = deserialize(tmp[1]['value'])
            code_dict = dict()
            for arg in args:
                arg_val = args[arg]
                if arg != '__doc__':
                    code_dict[arg] = arg_val
            code_list = [0] * 16

            for name in code_dict:
                if name == 'co_lnotab':
                    continue
                code_list[CODE_ARGS.index(name)] = code_dict[name]

            return types.CodeType(*code_list)


def deserialize_class(item):
    class_dict = deserialize(item[VALUE])
    name = class_dict[NAME]
    del class_dict[NAME]

    return type(name, (object,), class_dict)


def deserialize_module(item):
    return __import__(item[VALUE])


def deserialize_object(item):
    value = deserialize(item[VALUE])
    result = value[OBJECT_TYPE](**value[FIELDS])

    for key, value in value[FIELDS].items():
        result.key = value

    return result


# simple types
a = 465
aa = -46
b = 9.859
bb = -0.452
c = '935'
cc = ''
d = True
dd = False
e = None
f = complex(5, 8)

# collection types
list1 = [1, 2, 3, 4, 5]
list2 = ['1', 2]
list3 = [complex(2, 3), 174, '1', False, 0.56, None]

set1 = {1, 2, 3, 4, 5}
set2 = {'1', 2, 7}
set3 = {complex(2, 3), 1, '1', True, 0.9, None}

tuple1 = (1, 2, 3, 4, 5)
tuple2 = (1, '2', 46)
tuple3 = (complex(2, 3), 1, '1', True, 0.9, None)

bytes1 = bytes([46, 46, 47])

dict1 = {1: 2, 2: 4}
dict2 = {1: '1', 2: '2'}

dict3 = {1: [1, 2], 2: (1, 2), 3: '3', 4: False}
dict4 = {1: [1, 2, (3, 4, {5, 6, '7', False})], 2: (1, (2, (3, (4, (5, (6, (7)))))))}
dict5 = {(1, (2, (3, (4, (5, (6)))))): (7, (8, (9, (10, (12))))), \
         (1, ('2', (True, (None, (complex(2, 3), (6.9)))))): (1, (1, (1, (2, {2, (2, (5,))}))))}


# funcs

def func1():
    return 12


def func2(param):
    return 6 + param


def func3(*params):
    sum = 0

    for tmp in params:
        sum = sum + tmp

    return sum


def func4(param):
    import math

    return math.sin(param) + math.cos(param)


def func5(arr):
    return sorted(arr)


def func6(n):
    if (n == 1):
        return 1

    return n * func6(n - 1)


def func7(n):
    return n * func6(n - 1)


lambda1 = lambda a: a + 10
lambda2 = lambda a, b, c: a + b + c


# class test

class A:
    a = 123

    def __init__(self):
        pass

    def qwe(self, a):
        return a


class B(A):
    c = 123

    def __init__(self, b):
        self.b = b

    def func(self):
        return 8


class E:
    e = 123

    def __init__(self):
        pass


class EE(E):
    ee = 123

    def __init__(self):
        pass


class EEE(EE):
    eee = 123

    def __init__(self):
        pass


class EEEE(EEE):
    eeee = 123

    def __init__(self):
        pass


class first:
    def __init__(self):
        pass

    def func(self, a):
        return a


class second(first):
    def __init__(self):
        pass

    def func(self, a):
        return a ** a


class st:
    @staticmethod
    def qwe():
        return 5


class HardBase:
    a = 123
    s = '123'

    def __init__(self):
        pass

    def func(self, a, b):
        return a ** b


class HardClass(HardBase):
    c = 456

    def __init__(self):
        pass

    def func1(self, *args):
        import math
        sum = 0

        for tmp in args:
            sum += tmp

        return tmp

    def func2(self, a):
        return func7(a)


import types
import regex

from serializers.xml.xml_constants import *


def dumps_from_dict(obj, is_first=False) -> str:
    if type(obj) in (int, float, bool, types.NoneType):
        return create_xml_element(type(obj).__name__, str(obj), is_first)
    elif type(obj) is str:
        data = mask_symbols(obj)
        return create_xml_element(str.__name__, data, is_first)
    elif type(obj) is list:
        data = ''.join([dumps_from_dict(o) for o in obj])
        return create_xml_element(list.__name__, data, is_first)
    elif type(obj) is dict:
        data = ''.join(
            [f"{dumps_from_dict(item[0])}{dumps_from_dict(item[1])}" for item in obj.items()])
        return create_xml_element(dict.__name__, data, is_first)


def loads_to_dict(string: str, is_first=False):
    string = string.strip()
    xml_element_pattern = FIRST_XML_ELEMENT_PATTERN if is_first else XML_ELEMENT_PATTERN

    match = regex.fullmatch(xml_element_pattern, string)

    if not match:
        raise ValueError

    key = match.group(KEY_GROUP_NAME)
    value = match.group(VALUE_GROUP_NAME)

    if key == int.__name__:
        return int(value)
    elif key == float.__name__:
        return float(value)
    elif key == bool.__name__:
        return value == str(True)
    elif key == str.__name__:
        return unmask_symbols(value)
    elif key == types.NoneType.__name__:
        return None
    elif key == list.__name__:
        matches = regex.findall(XML_ELEMENT_PATTERN, value)
        return [loads_to_dict(match[0]) for match in matches]
    elif key == dict.__name__:
        matches = regex.findall(XML_ELEMENT_PATTERN, value)
        return {loads_to_dict(matches[i][0]):
                    loads_to_dict(matches[i + 1][0]) for i in range(0, len(matches), 2)}


def create_xml_element(name: str, data: str, is_first=False):
    if is_first:
        return f"<{name} {XML_SCHEME_PATTERN}>{data}</{name}>"
    else:
        return f"<{name}>{data}</{name}>"


def mask_symbols(string: str) -> str:
    return string.replace('&', "&amp;").replace('<', "&lt;").replace('>', "&gt;"). \
        replace('"', "&quot;").replace("'", "&apos;")


def unmask_symbols(string: str) -> str:
    return string.replace("&amp;", '&').replace("&lt;", '<').replace("&gt;", '>'). \
        replace("&quot;", '"').replace("&apos;", "'")


from serializers.json.json_serializer import JsonSerializer
from serializers.xml.xml_serializer import XmlSerializer


class Factory:
    @staticmethod
    def create_serializer(file_type):
        if file_type == "json":
            return JsonSerializer
        elif file_type == "xml":
            return XmlSerializer


KEY_GROUP_NAME = "key"
VALUE_GROUP_NAME = "value"
ELEMENTARY_NAMES_PATTERN = "int|float|bool|str|NoneType|list|dict"

XML_SCHEME_PATTERN = "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " + \
                     "xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\""

XML_ELEMENT_PATTERN = fr"(\<(?P<{KEY_GROUP_NAME}>{ELEMENTARY_NAMES_PATTERN})\>" + \
                      fr"(?P<{VALUE_GROUP_NAME}>([^<>]*)|(?R)+)\</(?:{ELEMENTARY_NAMES_PATTERN})\>)"

FIRST_XML_ELEMENT_PATTERN = fr"(\<(?P<{KEY_GROUP_NAME}>{ELEMENTARY_NAMES_PATTERN})\s*({XML_SCHEME_PATTERN})?\>" + \
                            fr"(?P<{VALUE_GROUP_NAME}>([^<>]*)|(?R)+)\</(?:{ELEMENTARY_NAMES_PATTERN})\>)"


from serializers.source.serializer import serialize, deserialize
from test_variables import *


def test_simple_1():
    assert a == deserialize(serialize(a))


def test_simple_2():
    assert aa == deserialize(serialize(aa))


def test_simple_3():
    assert b == deserialize(serialize(b))


def test_simple_4():
    assert bb == deserialize(serialize(bb))


def test_simple_5():
    assert c == deserialize(serialize(c))


def test_simple_6():
    assert cc == deserialize(serialize(cc))


def test_simple_7():
    assert d == deserialize(serialize(d))


def test_simple_8():
    assert dd == deserialize(serialize(dd))


def test_simple_9():
    assert e == deserialize(serialize(e))


def test_simple_10():
    assert f == deserialize(serialize(f))


def test_collection_1():
    assert list1 == deserialize(serialize(list1))


def test_collection_2():
    assert list2 == deserialize(serialize(list2))


def test_collection_3():
    assert list3 == deserialize(serialize(list3))


def test_collection_4():
    assert set1 == deserialize(serialize(set1))


def test_collection_5():
    assert set2 == deserialize(serialize(set2))


def test_collection_6():
    assert set3 == deserialize(serialize(set3))


def test_collection_7():
    assert tuple1 == deserialize(serialize(tuple1))


def test_collection_8():
    assert tuple2 == deserialize(serialize(tuple2))


def test_collection_9():
    assert tuple3 == deserialize(serialize(tuple3))


def test_collection_10():
    assert bytes1 == deserialize(serialize(bytes1))


def test_collection_11():
    assert dict1 == deserialize(serialize(dict1))


def test_collection_12():
    assert dict2 == deserialize(serialize(dict2))


def test_collection_13():
    assert dict3 == deserialize(serialize(dict3))


def test_collection_14():
    assert dict4 == deserialize(serialize(dict4))


def test_collection_15():
    assert dict5 == deserialize(serialize(dict5))


def test_func_1():
    assert func1() == deserialize(serialize(func1))()


def test_func_2():
    assert func2(12) == deserialize(serialize(func2))(12)


def test_func_3():
    assert func3(12) == deserialize(serialize(func3))(12)


def test_func_4():
    assert func3(12, 12, 12, 12, 12) == deserialize(serialize(func3))(12, 12, 12, 12, 12)


def test_func_5():
    assert func4(0.5) == deserialize(serialize(func4))(0.5)


def test_func_6():
    assert func5([4, 3, 2, 1]) == deserialize(serialize(func5))([4, 3, 2, 1])


def test_func_7():
    assert func6(5) == deserialize(serialize(func6))(5)


def test_func_8():
    assert func7(5) == deserialize(serialize(func7))(5)


def test_func_9():
    tmp = deserialize(serialize(lambda1))

    assert tmp(5) == lambda1(5)


def test_func_10():
    tmp = deserialize(serialize(lambda2))

    assert tmp(1, 2, 3) == lambda2(3, 2, 1)


def test_class_1():
    tmp = deserialize(serialize(A))
    tmp = tmp()
    a = A()

    assert a.a == tmp.a
    assert a.qwe(2) == tmp.qwe(2)


def test_class_2():
    tmp = deserialize(serialize(B))
    tmp = tmp(2)
    b = B(2)

    assert tmp.a == b.a
    assert tmp.func() == b.func()
    assert tmp.qwe(4) == b.qwe(4)


def test_class_3():
    tmp = deserialize(serialize(EEEE))
    tmp = tmp()
    eeee = EEEE()

    assert tmp.e == eeee.e
    assert tmp.ee == eeee.ee
    assert tmp.eee == eeee.eee
    assert tmp.eeee == eeee.eeee


def test_class_4():
    tmp = deserialize(serialize(first))
    f = tmp()
    fo = first()

    tmp = deserialize(serialize(second))
    s = tmp()
    so = second()

    assert s.func(4) == so.func(4)
    assert f.func(4) ** 4 == s.func(4)


def test_class_object_1():
    a = A()
    tmp = deserialize(serialize(a))

    assert tmp.a == a.a
    assert tmp.qwe(4) == a.qwe(4)


def test_class_5():
    tmp = deserialize(serialize(st))

    assert tmp.qwe() == st.qwe()


def test_class_6():
    tmp = deserialize(serialize(HardClass))

    h = HardClass()
    tmp = tmp()

    assert tmp.a == h.a
    assert tmp.s == h.s
    assert tmp.func(1, 2) == h.func(1, 2)

    assert tmp.c == h.c
    assert tmp.func1(1, 2, 3, 4) == h.func1(1, 2, 3, 4)
    assert tmp.func2(5) == h.func2(5)


test_simple_1()
test_simple_2()
test_simple_3()
test_simple_4()
test_simple_5()
test_simple_6()
test_simple_7()
test_simple_8()
test_simple_9()
test_simple_10()

test_collection_1()
test_collection_2()
test_collection_3()
test_collection_4()
test_collection_5()
test_collection_6()
test_collection_7()
test_collection_8()
test_collection_9()
test_collection_10()
test_collection_11()
test_collection_12()
test_collection_13()
test_collection_14()
test_collection_15()

test_func_1()
test_func_2()
test_func_3()
test_func_4()
test_func_5()
test_func_6()
test_func_7()
test_func_8()
test_func_9()
test_func_10()

test_class_1()
test_class_2()
test_class_3()
test_class_4()
test_class_5()
test_class_6()

test_class_object_1()


from serializers.json.json_parser import parse_json
from serializers.source.serializer import serialize, deserialize


class JsonSerializer:
    @staticmethod
    def dumps(item):
        return str(serialize(item)).replace("\n", "\\n")

    @staticmethod
    def dump(item, file):
        file.write(JsonSerializer.dumps(item))

    @staticmethod
    def loads(item):
        return deserialize(parse_json(item.replace("\\n", "\n")))

    @staticmethod
    def load(file):
        return JsonSerializer.loads(file.read())


from serializers.xml.xml_parser import dumps_from_dict, loads_to_dict
from serializers.source.serializer import serialize, deserialize


class XmlSerializer:
    @staticmethod
    def dumps(obj):
        obj = serialize(obj)

        return dumps_from_dict(obj)

    @staticmethod
    def loads(item):
        item = loads_to_dict(item)

        return deserialize(item)

    @staticmethod
    def dump(item, file):
        file.write(XmlSerializer.dumps(item))

    @staticmethod
    def load(file):
        return XmlSerializer.loads(file.read())


from setuptools import setup

setup(
    name="Gergyg_lab3_json_xml_serializer",
    version="0.1",
    description="package for python (de)serialization in .json and .xml",
    url="https://github.com/Gergyg/Python-labs/tree/Lab3",
    author="Arseni Matsiusheuski",
    author_email="gad.olg@mail.ru",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent"
    ],
    packages=["serializers/json", "serializers/source", "serializers/xml", "serializers", "tests"],
    include_package_data=True
)


def parse_json(item):
    if item == "[]":
        return {}
    elif item[0] == "{" or item[0] == "[":
        is_list = False
        if item[0] == "[":
            is_list = True

        item = item[1:len(item) - 1]
        result = {}
        result_list = []
        depth = 0
        quote = False
        temp_str = ""
        key = ""
        value = ""

        for char in item:
            if char == "{":
                depth += 1
            elif char == "}":
                depth -= 1
            elif char == "[":
                depth += 1
            elif char == "]":
                depth -= 1
            elif char == "\"":
                quote = not quote
            elif char == ":" and not quote and depth == 0:
                key += temp_str[1:len(temp_str) - 1]
                temp_str = ""
                continue
            elif char == "," and not quote and depth == 0 and not is_list:
                value += temp_str[1:len(temp_str) - 1]
                result[key] = value
                value = ""
                key = ""
                temp_str = ""
                continue
            elif char == "," and is_list and depth == 0:
                result_list.append(parse_json(temp_str))
                temp_str = ""
                continue
            elif char == ' ' and not quote:
                continue

            temp_str += char

        if temp_str[0] == "'" or temp_str[0] == '"':
            result[key] = temp_str[1:len(temp_str) - 1]
        elif is_list:
            result_list.append(parse_json(temp_str))
        else:
            result[key] = parse_json(temp_str)
    else:
        return item

    if is_list:
        return result_list
    else:
        return result